import logging
import os
from contextlib import redirect_stdout
from typing import Iterator, Optional, TextIO

import hojichar

logger = logging.getLogger(__name__)


def process_iter(
    input_iter: Iterator[str],
    filter: hojichar.Compose,
    exit_on_error: bool,
    stdout_fp: Optional[TextIO] = None,
) -> Iterator[hojichar.Document]:
    """
    Getting an iterator of string, processing by given hojichar.Compose filter,
    and iterate processed string.
    Standard output of each filter is redirected to stdout_fp. Default is /dev/null.

    Args:
        input_iter (Iterator[str]): Input iterator.
        filter (hojichar.Compose): Processing filter.
        exit_on_error (bool): Halt with error while processing.
        stdout_fp: Redirection of stdout in each filter.

    Raises:
        e: Exception raised during processing in hojichar.Compose

    Yields:
        Iterator[hojichar.Document]: Processed text
    """
    if stdout_fp is None:
        stdout_fp = open(os.devnull, "w")
    with redirect_stdout(stdout_fp):
        for line in input_iter:
            try:
                doc = filter.apply(hojichar.Document(line))
                yield doc
            except Exception as e:
                if exit_on_error:
                    raise e
                else:
                    logger.error(f"Caught {type(e)}. Skip processing the line: `{line}`")
                    continue


def reject_iter(input_iter: Iterator[hojichar.Document], discard_rejected: bool) -> Iterator[str]:
    """_summary_

    Args:
        input_iter (Iterator[hojichar.Document]): Input iterator of the documents.
        discard_rejected (bool): Whether discarding `is_rejected` docs or not.

    Yields:
        Iterator[str]: Output iterator of the string.
    """
    for doc in input_iter:
        if doc.is_rejected and discard_rejected:
            pass
        else:
            yield doc.text



from __future__ import annotations

import dataclasses
import logging
import time
from typing import Any, Dict, List, Union

from hojichar.core.filter_interface import Filter, TokenFilter
from hojichar.core.models import Document

logger = logging.getLogger(__name__)


class Inspector(Filter):
    def __init__(
        self, target_filter: Union[Filter, TokenFilter], filter_idx: int, *args: Any, **kwargs: Any
    ) -> None:
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger("hojichar.Inspector")
        self.target_filter = target_filter
        self.filter_idx = filter_idx
        self.target = f"{filter_idx}-{target_filter.name}"

        self.is_rejected = False
        self.text_hash = 0
        self.tokens_hash = 0

    def apply(self, document: Document) -> Document:
        self.inspect(document)
        return document

    def inspect(self, document: Document) -> None:
        self.is_rejected = False
        self.is_rejected = document.is_rejected
        self.bytes = len(document.text.encode("utf-8"))
        self.time_ns = time.perf_counter_ns()


@dataclasses.dataclass
class FilterStatistics:
    name: str
    discard_num: int = 0
    diff_bytes: int = 0
    cumulative_time_ns: int = 0
    params: Dict[str, Any] = dataclasses.field(default_factory=dict)

    def get_human_readable_values(self) -> dict:
        ret = {
            "name": self.name,
            "discard_num": self.discard_num,
            "diff_MB": (self.diff_bytes / 1048576),  # 1024**2
            "cumulative_time": (self.cumulative_time_ns / 10**9),
            "params": self.params,
        }
        return ret

    def __add__(self, other: FilterStatistics) -> FilterStatistics:
        assert self.name == other.name, "Layer names must match"
        return FilterStatistics(
            self.name,
            self.discard_num + other.discard_num,
            self.diff_bytes + other.diff_bytes,
            self.cumulative_time_ns + other.cumulative_time_ns,
            self.params,
        )


@dataclasses.dataclass
class DocStatistics:
    processed_num: int = 0
    discard_num: int = 0
    input_bytes: int = 0
    output_bytes: int = 0
    cumulative_time_ns: int = 0
    total_token_num: int = 0

    def get_human_readable_values(self) -> dict:
        ret = {
            "processed_num": self.processed_num,
            "discard_num": self.discard_num,
            "input_MB": (self.input_bytes / 1000**2),
            "output_MB": (self.output_bytes / 1000**2),
            "cumulative_time": (self.cumulative_time_ns / 10**9),
            "total_token_num": self.total_token_num,
        }
        return ret

    def __add__(self, other: DocStatistics) -> DocStatistics:
        return DocStatistics(
            self.processed_num + other.processed_num,
            self.discard_num + other.discard_num,
            self.input_bytes + other.input_bytes,
            self.output_bytes + other.output_bytes,
            self.cumulative_time_ns + other.cumulative_time_ns,
            self.total_token_num + other.total_token_num,
        )


@dataclasses.dataclass
class StatsContainer:
    total_info: DocStatistics
    layers_info: Dict[str, FilterStatistics]  # Key of the dict is filter name.

    def __add__(self, other: StatsContainer) -> StatsContainer:
        assert self.layers_info.keys() == other.layers_info.keys(), "Layer names must match"
        return StatsContainer(
            self.total_info + other.total_info,
            {k: v + other.layers_info[k] for k, v in self.layers_info.items()},
        )

    def get_human_readable_values(self) -> dict:
        return {
            "total_info": self.total_info.get_human_readable_values(),
            "layers_info": [
                layer.get_human_readable_values() for layer in self.layers_info.values()
            ],
        }


class StatisticsCounter:
    def __init__(self, inspectors: List[Inspector]) -> None:
        counts = dict()
        for inspector in inspectors:
            counts[inspector.target] = FilterStatistics(
                name=inspector.target,
                params=inspector.target_filter.get_jsonalbe_vars(),
            )
        self.stats = StatsContainer(
            DocStatistics(),
            counts,
        )

    def update_changes(
        self,
        document: Document,
        before_process_inspector: Inspector,
        inspectors: List[Inspector],
    ) -> None:

        # Counting statistics for each filter
        previous_inspector = before_process_inspector
        for idx, inspector in enumerate(inspectors):
            # Logging how many docs are discarded in each filter
            if (not previous_inspector.is_rejected) and inspector.is_rejected:
                self.stats.layers_info[inspector.target].discard_num += 1

            # logging how much volume of docs are changed in each filter.
            if (not previous_inspector.is_rejected) and inspector.is_rejected:
                diff_bytes = -inspector.bytes
            elif previous_inspector.is_rejected and inspector.is_rejected:
                diff_bytes = 0
            else:
                diff_bytes = inspector.bytes - previous_inspector.bytes

            self.stats.layers_info[inspector.target].diff_bytes += diff_bytes

            process_time_ns = inspector.time_ns - previous_inspector.time_ns
            self.stats.layers_info[inspector.target].cumulative_time_ns += process_time_ns

            previous_inspector = inspector

        # Counting total statistics
        self.stats.total_info.processed_num += 1
        self.stats.total_info.discard_num += (
            1 if any([inspector.is_rejected for inspector in inspectors]) > 0 else 0
        )
        self.stats.total_info.input_bytes += len(document.original.encode("utf-8"))
        self.stats.total_info.output_bytes += (
            0 if document.is_rejected else len(document.text.encode("utf-8"))
        )
        self.stats.total_info.cumulative_time_ns += inspectors[-1].time_ns - inspectors[0].time_ns
        self.stats.total_info.total_token_num += len(document.tokens)

    def get_statistics(self) -> dict:
        return self.stats.get_human_readable_values()



from typing import Any, Dict, List, Optional


class Token:
    def __init__(self, text: str, is_rejected: bool = False) -> None:
        self.text = text
        self.__original = text
        self.is_rejected = is_rejected

    @property
    def original(self) -> str:
        return self.__original

    def __str__(self) -> str:
        return self.text


class Document:
    def __init__(
        self, text: str, is_rejected: bool = False, tokens: Optional[List[Token]] = None
    ) -> None:
        self.text = text
        self.__original = text
        self.is_rejected = is_rejected
        if tokens is None:
            self.tokens: List[Token] = []

        self.dedup_lsh: List[str] = []
        self.reject_reason: Dict[str, Any] = {}

    @property
    def original(self) -> str:
        return self.__original

    def set_tokens(self, tokens: List[str]) -> None:
        self.tokens = [Token(token) for token in tokens]

    def get_tokens(self) -> List[str]:
        return [token.text for token in self.tokens]

    def __str__(self) -> str:
        return self.text



import argparse
import functools
import json
import logging
import multiprocessing
import os
import sys
from typing import Dict, Iterator, Tuple

import hojichar
from hojichar.core.inspection import StatsContainer
from hojichar.utils.io_iter import fileout_from_iter, stdin_iter, stdout_from_iter
from hojichar.utils.load_compose import load_compose

MAIN_FILTER: hojichar.Compose
CLI_ARGS: argparse.Namespace

logger = logging.getLogger("hojichar.cli")


def argparser() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--profile",
        "-p",
        required=True,
        metavar="<profile.py>",
        help="Path to a Python file that implements your custom filter.",
    )
    parser.add_argument(
        "--args",
        default=[],
        nargs="+",
        help="Pass additional arguments to the profile.\
            Use it like `--args arg1 arg2` etc. The arguments should be space-separated.",
    )
    parser.add_argument(
        "--output",
        "-o",
        default=None,
        help="Specifies the path for the output file. Defaults to standard output.",
    )
    parser.add_argument(
        "--dump-stats",
        default=None,
        metavar="<path to stats.json>",
        help="Dump statistics to file. If the file exists, it will be appended.",
    )
    parser.add_argument(
        "--exit-on-error",
        action="store_true",
        help="Exit if an exception occurs during filtering.\
            Useful for debugging custom filters.",
    )
    parser.add_argument(
        "--all",
        action="store_true",
        default=False,
        help="A flag that specifies whether to include discarded samples. \
            This is useful when inspecting discarded samples.",
    )
    parser.add_argument(
        "--jobs",
        "-j",
        default=None,
        type=int,
        help="The number ob parallel jobs. By default, the nuber of the CPU core.",
    )
    args = parser.parse_args()
    return args


def init_worker(filter: hojichar.Compose, args: argparse.Namespace) -> None:
    global MAIN_FILTER, CLI_ARGS
    MAIN_FILTER = filter
    CLI_ARGS = args


def worker(doc: hojichar.Document) -> Tuple[hojichar.Document, int, StatsContainer]:
    global MAIN_FILTER, CLI_ARGS
    try:
        return MAIN_FILTER.apply(doc), os.getpid(), MAIN_FILTER.statistics_obj
    except Exception as e:
        if CLI_ARGS.exit_on_error:
            raise e
        else:
            logger.error(f"Caught {type(e)}. Skip processing the line: {doc.text}")
            return hojichar.Document("", is_rejected=True), os.getpid(), MAIN_FILTER.statistics_obj


def out_doc_generator(
    worker_out_iter: Iterator[Tuple[hojichar.Document, int, StatsContainer]],
    pid_stats: Dict[int, StatsContainer],
) -> Iterator[hojichar.Document]:
    for worker_out in worker_out_iter:
        doc, pid, worker_stats = worker_out
        pid_stats[pid] = worker_stats
        yield doc


def main() -> None:
    pid_stats: Dict[int, StatsContainer] = dict()

    args = argparser()
    input_iter = stdin_iter()
    input_doc_iter = (hojichar.Document(s) for s in input_iter)
    filter = load_compose(args.profile, *tuple(args.args))
    with multiprocessing.Pool(
        processes=args.jobs, initializer=init_worker, initargs=(filter, args)
    ) as pool:
        worker_out_iter = pool.imap_unordered(worker, input_doc_iter)
        out_doc_iter = out_doc_generator(worker_out_iter, pid_stats)
        out_str_iter = (
            (doc.text for doc in out_doc_iter)
            if args.all
            else (doc.text for doc in out_doc_iter if not doc.is_rejected)
        )
        if args.output:
            with open(args.output, "w") as fp:
                fileout_from_iter(out_str_iter, fp)
        else:
            stdout_from_iter(out_str_iter)

    stats: StatsContainer = functools.reduce(lambda x, y: x + y, pid_stats.values())
    print(
        json.dumps(stats.get_human_readable_values(), ensure_ascii=False, indent=2),
        file=sys.stderr,
    )
    if args.dump_stats:
        with open(args.dump_stats, "a") as fp:
            fp.write(json.dumps(stats.get_human_readable_values(), ensure_ascii=False) + "\n")


if __name__ == "__main__":
    main()
